# pylint: disable=missing-docstring

from polarion_docstrings.parser import (
    DocstringRecord,
    ValueRecord,
    get_docstrings_in_file,
    merge_docstrings,
    merge_record,
)

EXPECTED = {
    "tests/data/docstrings_merge.py": {"assignee": "mkourim", "foo": "this is an unknown field"},
    "tests/data/docstrings_merge.py::TestClassFoo": {
        "assignee": "psegedy",
        "caseimportance": "huge",
        "bar": "this is an unknown field",
        "linkedWorkItems": "FOO, BAR",
        "testSteps": [
            "1. Step with really long description that doesn't fit into one line",
            "2. Do that",
        ],
    },
    "tests/data/docstrings_merge.py::TestClassFoo::test_in_class_no_docstring": {
        "assignee": "psegedy",
        "foo": "this is an unknown field",
        "caseimportance": "huge",
        "bar": "this is an unknown field",
        "linkedWorkItems": "FOO, BAR",
        "testSteps": [
            "1. Step with really long description that doesn't fit into one line",
            "2. Do that",
        ],
    },
    "tests/data/docstrings_merge.py::TestClassFoo::test_in_class_no_polarion": {
        "assignee": "psegedy",
        "foo": "this is an unknown field",
        "caseimportance": "huge",
        "bar": "this is an unknown field",
        "linkedWorkItems": "FOO, BAR",
        "testSteps": [
            "1. Step with really long description that doesn't fit into one line",
            "2. Do that",
        ],
    },
    "tests/data/docstrings_merge.py::TestClassFoo::test_in_class_polarion": {
        "assignee": "mkourim",
        "foo": "this is an unknown field",
        "caseimportance": "low",
        "bar": "this is an unknown field",
        "linkedWorkItems": "FOO, BAR",
        "testSteps": ["1. Do that"],
        "baz": "this is an unknown field",
    },
    "tests/data/docstrings_merge.py::test_standalone_no_docstring": {
        "assignee": "mkourim",
        "foo": "this is an unknown field",
    },
}

TO_MERGE = (
    "tests/data/docstrings_merge.py::TestClassFoo::test_record_merge",
    DocstringRecord(
        lineno=72,
        column=4,
        value={
            "assignee": ValueRecord(lineno=1, column=8, value="bender"),
            "testSteps": ValueRecord(lineno=3, column=8, value="wrong"),
        },
        nodeid="tests/data/docstrings_merge.py::TestClassFoo::test_record_merge",
        level="function",
    ),
)

EXPECTED_MERGED_RECORD = {
    "assignee": "bender",
    "foo": "this is an unknown field",
    "caseimportance": "huge",
    "bar": "this is an unknown field",
    "linkedWorkItems": "FOO, BAR",
    "testSteps": "wrong",
}

EXPECTED_MERGED_EMPTY_RECORD = {
    "assignee": "psegedy",
    "foo": "this is an unknown field",
    "caseimportance": "huge",
    "bar": "this is an unknown field",
    "linkedWorkItems": "FOO, BAR",
    "testSteps": [
        "1. Step with really long description that doesn't fit into one line",
        "2. Do that",
    ],
}


def test_parser_merged(source_file_merge):
    docstrings = get_docstrings_in_file(source_file_merge)
    merged_docstrings = merge_docstrings(docstrings)
    assert merged_docstrings == EXPECTED


def test_parser_docstrings_record_merge(source_file_merge):
    docstrings = get_docstrings_in_file(source_file_merge)
    merged_docstrings = merge_docstrings(docstrings)
    merged_record = merge_record(TO_MERGE[1], merged_docstrings)
    assert merged_record == EXPECTED_MERGED_RECORD


def test_parser_value_merge(source_file_merge):
    docstrings = get_docstrings_in_file(source_file_merge)
    merged_docstrings = merge_docstrings(docstrings)
    merged_record = merge_record(TO_MERGE[1].value, merged_docstrings, TO_MERGE[0])
    assert merged_record == EXPECTED_MERGED_RECORD


def test_parser_empty_merge(source_file_merge):
    docstrings = get_docstrings_in_file(source_file_merge)
    merged_docstrings = merge_docstrings(docstrings)
    merged_record = merge_record({}, merged_docstrings, TO_MERGE[0])
    assert merged_record == EXPECTED_MERGED_EMPTY_RECORD



# pylint: disable=missing-docstring,no-self-use

import pytest


class TestClassFoo:
    def test_in_class_no_docstring(self):
        pass

    def test_in_class_no_polarion(self):
        """FOO"""

    @pytest.mark.skip
    def test_in_class_polarion(self):
        """FOO

        Polarion:
            assignee: mkourim
            casecomponent: nonexistent
            caseimportance: low

            testSteps:
                1. Step with really long description
                   that doesn't fit into one line
                2. Do that

            expectedResults:
                1. Success outcome with really long description
                   that doesn't fit into one line
                2. second

            title: Some test with really long description
                   that doesn't fit into one line

            setup: Do this:
                   - first thing
                   - second thing

            teardown: Tear it down
            caselevel: level1
            caseautomation: automated
            linkedWorkItems: FOO, BAR
            foo: this is an unknown field
            description: ignored

        This is not included.
        """


@pytest.mark.skip
def test_annotated_no_docstring():
    pass


def test_standalone_no_docstring():
    pass


@pytest.mark.skip
def test_annotated_no_polarion():
    """FOO"""


@pytest.mark.skip
def test_annotated_polarion():

    """FOO

    Polarion:
        assignee: mkourim
        initialEstimate: 1/4h
        testSteps: wrong


        expectedResults:
    """


def test_blacklisted():
    """FOO

    Polarion:
        initialEstimate: 1/4h
    """


def test_blacklisted_and_whitelisted():
    """FOO

    Polarion:
        initialEstimate: 1/4
            assignee: someone
    """


def test_wrong_steps_indent():
    """Polarion:
        initialEstimate: 1/4m
        testSteps:
            1. Step with really long description
               that doesn't fit into one line
               assignee: someone"""


def test_multiline():
    """BAR
    Polarion:
        assignee: mkourim
        initialEstimate: 1/4h
        title: >
            Test multiple tenant
            quotas simultaneously
        testSteps:
            1.
            2. >
               Log something
        expectedResults:
            1.
            2. >
               <NoMethodError>: <undefined method `service_resources" for
               nil:NilClass> [----] E, [2018-01-06T11:11:20.074019 #13027:e0ffc4] ERROR -- :"
    """



# pylint: disable=missing-docstring

import os

import pytest
import yaml

TESTS_DIR = os.path.dirname(os.path.relpath(__file__))
TESTS_DATA_DIR = os.path.join(TESTS_DIR, "data")
CONFIG_TEMPLATE = os.path.join(TESTS_DIR, os.pardir, "polarion_tools.yaml.template")


@pytest.fixture(scope="session")
def source_file():
    return os.path.join(TESTS_DATA_DIR, "docstrings_func.py")


@pytest.fixture(scope="session")
def source_file_merge():
    return os.path.join(TESTS_DATA_DIR, "docstrings_merge.py")


@pytest.fixture(scope="session")
def config():
    with open(CONFIG_TEMPLATE, encoding="utf-8") as config_file:
        config_loaded = yaml.safe_load(config_file)
    return config_loaded



# pylint: disable=missing-docstring

from polarion_docstrings import checker

EXPECTED = [
    (6, 4, 'P666 Unknown field "foo"'),
    (17, 8, 'P667 Invalid value "huge" of the "caseimportance" field'),
    (18, 8, 'P666 Unknown field "bar"'),
    (
        19,
        8,
        'P668 Field "linkedWorkItems" should be handled by the "@pytest.mark.requirements" marker',
    ),
    (20, 12, "P663 Wrong indentation, line ignored"),
    (28, 4, 'P669 Missing required field "initialEstimate"'),
    (32, 8, 'P669 Missing required field "initialEstimate"'),
    (38, 8, 'P669 Missing required field "initialEstimate"'),
    (41, 12, 'P666 Unknown field "baz"'),
    (48, 0, 'P669 Missing required field "initialEstimate"'),
]


def _strip_func(errors):
    return [(lineno, col, msg) for lineno, col, msg, __ in errors]


def test_checker(source_file_merge, config):
    errors = checker.DocstringsChecker(None, source_file_merge, config, "TestChecker").run_checks()
    errors = _strip_func(errors)
    assert errors == EXPECTED



"""
Parses Polarion docstrings.
"""

import ast
from collections import namedtuple

FORMATED_KEYS = ("setup", "teardown")

DocstringRecord = namedtuple("DocstringRecord", "lineno column value nodeid level")
ValueRecord = namedtuple("ValueRecord", "lineno column value")
ErrorRecord = namedtuple("ErrorRecord", "lineno column type message")


def get_section_start(doc_list, section):
    """Finds the line with "section" (e.g. "Polarion", "testSteps", etc.)."""
    section = "{}:".format(section)
    for index, line in enumerate(doc_list):
        if section != line.strip():
            continue
        indent = len(line) - len(line.lstrip(" "))
        return index + 1, indent
    return None, None


def add_errors(docstring_dict, errors):
    """Adds new `ErrorRecord`(s) to `_errors_parsing`."""
    if not errors:
        return

    if "_errors_parsing" not in docstring_dict:
        docstring_dict["_errors_parsing"] = []

    if isinstance(errors, ErrorRecord):
        docstring_dict["_errors_parsing"].append(errors)
    else:
        docstring_dict["_errors_parsing"].extend(errors)


# pylint: disable=too-few-public-methods,invalid-name
class LEVELS:  # noqa
    """Valid values for levels of Polarion docstrings."""

    file = "file"
    klass = "class"
    function = "function"


# pylint: disable=too-few-public-methods,invalid-name
class DOCSTRING_SECTIONS:  # noqa
    """Valid sections in Polarion docstring."""

    polarion = "Polarion"
    steps = "testSteps"
    results = "expectedResults"


class DocstringParser:
    """Parser for single docstring."""

    SECTIONS = DOCSTRING_SECTIONS

    def __init__(self, docstring):
        self.docstring = docstring

    @staticmethod
    def _get_key_value(line):
        """Gets the key and value out of docstring line."""
        data = line.split(":")
        if len(data) == 1:
            data.append("")

        key = data[0].strip()

        value = ":".join(data[1:]).strip()
        if value.lower() in ("none", "null"):
            value = None

        return key, value

    # pylint: disable=too-many-locals
    def lines_to_dict(self, lines, start=0, lineno_offset=0, stop=None):
        """Creates dictionary out of docstring lines.

        Includes column and line number info for each record.
        """
        if stop:
            lines = lines[start:stop]
        else:
            lines = lines[start:]

        docstring_dict = {}
        indent = len(lines[0]) - len(lines[0].lstrip(" "))
        is_multiline = False
        prev_key = None
        for num, line in enumerate(lines, 1):
            line_stripped = line.strip()
            if not line_stripped:
                continue

            curr_indent = len(line) - len(line.lstrip(" "))

            if curr_indent < indent:
                break

            # check if the line should be appended to previous key
            is_first_multiline = False
            if not is_multiline:
                is_first_multiline = line_stripped[-1] == ">"
                is_first_multiline = is_first_multiline and len(line_stripped.split(" ")) <= 2
                if is_first_multiline:
                    # we might want to append this very line to `prev_key` if it's in a list
                    is_multiline = True
                    # remove the ">"
                    line_stripped = line_stripped[:-1].strip()

            first_word = line_stripped.split(" ")[0] or line_stripped
            is_keyword = first_word[-1] == ":"
            if prev_key and curr_indent > indent and (not is_keyword or is_multiline):
                prev_lineno, prev_indent, prev_value = docstring_dict[prev_key]
                sep = ""
                if prev_value:
                    sep = "\n" if prev_key in FORMATED_KEYS else " "
                docstring_dict[prev_key] = ValueRecord(
                    prev_lineno, prev_indent, "{}{}{}".format(prev_value, sep, line_stripped)
                )
                continue

            prev_key = None
            # if we are here we are either at the very beginning of multiline
            # (and `is_first_multiline` is True) or it's no longer multiline
            is_multiline = is_first_multiline

            if curr_indent > indent:
                add_errors(
                    docstring_dict,
                    ErrorRecord(
                        num + lineno_offset,
                        curr_indent,
                        "indent",
                        "Wrong indentation, line ignored",
                    ),
                )
                continue

            key, value = self._get_key_value(line_stripped)
            docstring_dict[key] = ValueRecord(num + lineno_offset, indent, value.strip())
            prev_key = key

        return docstring_dict

    @staticmethod
    def lines_to_list(lines, start=0, lineno_offset=0, stop=None):
        """Creates list out of docstring lines.

        Includes column and line number info for each line.
        """
        if stop:
            lines = lines[start:stop]
        else:
            lines = lines[start:]

        lines_list = []
        is_multiline = False
        indent = len(lines[0]) - len(lines[0].lstrip(" "))
        for num, line in enumerate(lines, 1):
            line_stripped = line.strip()
            if not line_stripped:
                continue
            curr_indent = len(line) - len(line.lstrip(" "))
            if curr_indent < indent:
                break

            # check if the line should be appended to previous key
            is_first_multiline = False
            if not is_multiline:
                is_first_multiline = line_stripped[-1] == ">"
                is_first_multiline = is_first_multiline and len(line_stripped.split(" ")) <= 2
                if is_first_multiline:
                    # remove the ">"
                    line_stripped = line_stripped[:-1].strip()

            first_word = line_stripped.split(" ")[0] or line_stripped
            is_keyword = first_word[-1] == ":"
            if curr_indent > indent and (not is_keyword or is_multiline):
                prev_lineno, prev_indent, prev_value = lines_list.pop()
                lines_list.append(
                    ValueRecord(prev_lineno, prev_indent, "{} {}".format(prev_value, line_stripped))
                )
                continue

            # if we are here we are either at the very beginning of multiline
            # (and `is_first_multiline` is True) or it's no longer multiline
            is_multiline = is_first_multiline

            if curr_indent > indent:
                continue
            lines_list.append(ValueRecord(num + lineno_offset, indent, line_stripped))

        return lines_list

    def strip_polarion_data(self):
        """Removes the Polarion section from the docstring."""
        docstring_list = self.docstring.split("\n")
        new_docstring_list = []
        polarion_section = "{}:".format(self.SECTIONS.polarion)
        indent = 0
        in_polarion = False

        for line in docstring_list:
            if line.strip() == polarion_section:
                indent = len(line) - len(line.lstrip(" "))
                in_polarion = True
                continue
            if indent:
                curr_indent = len(line) - len(line.lstrip(" "))
                if curr_indent > indent:
                    continue
                # If this is the next line after Polarion section,
                # don't save blank line.
                if in_polarion:
                    in_polarion = False
                    if not line.strip():
                        continue
            new_docstring_list.append(line)

        new_docstring = "\n".join(new_docstring_list)
        return new_docstring.strip()

    def parse(self):
        """Parses docstring to dictionary. E.g.:

        Polarion:
            assignee: mkourim
            casecomponent: nonexistent
            testSteps:
                1. Step with really long description
                   that doesn't fit into one line
                2. Do that
            expectedResults:
                1. Success outcome with really long description
                   that doesn't fit into one line
                2. 2
            caseimportance: low
            title: Some test with really long description
                   that doesn't fit into one line
            setup: Do this:
                   - first thing
                   - second thing

        This is not included.
        """
        doc_list = self.docstring.split("\n")

        polarion_start, __ = get_section_start(doc_list, self.SECTIONS.polarion)
        if not polarion_start:
            return None

        docstring_dict = self.lines_to_dict(doc_list, start=polarion_start)

        if self.SECTIONS.steps in docstring_dict and docstring_dict[self.SECTIONS.steps].value:
            steps_start, __ = get_section_start(doc_list, self.SECTIONS.steps)
            if steps_start:
                steps_list = self.lines_to_list(
                    doc_list, start=steps_start, lineno_offset=steps_start - polarion_start
                )
                docstring_dict[self.SECTIONS.steps] = steps_list

        if self.SECTIONS.results in docstring_dict and docstring_dict[self.SECTIONS.results].value:
            results_start, __ = get_section_start(doc_list, self.SECTIONS.results)
            if results_start:
                results_list = self.lines_to_list(
                    doc_list, start=results_start, lineno_offset=results_start - polarion_start
                )
                docstring_dict[self.SECTIONS.results] = results_list

        return docstring_dict


class FileParser:
    """Parser for whole file."""

    SECTIONS = DOCSTRING_SECTIONS
    INDENT_STEP = 4

    def __init__(self, filename, tree=None, tests_only=True):
        self.filename = filename
        self.tests_only = tests_only
        self.tree = tree or self.get_tree()

    def get_tree(self):
        """Returns ast tree."""
        with open(self.filename) as infile:
            source = infile.read()

        tree = ast.parse(source, filename=self.filename)
        return tree

    @staticmethod
    def get_docstring_from_node(node):
        """Gets docstring from ast node."""
        docstring = None
        body = getattr(node, "body", None)
        body = body[0] if body else node

        try:
            if isinstance(body, ast.Expr) and isinstance(body.value, ast.Str):
                docstring = body.value.s
        # pylint: disable=broad-except
        except Exception:
            return None

        # for Python2
        try:
            docstring = docstring.decode("utf-8")
        except AttributeError:
            pass

        return docstring

    def _get_nodeid(self, node_name, class_name):
        components = [self.filename, class_name, node_name]
        if not class_name:
            components.pop(1)
        if not node_name:
            components.pop()
        nodeid = "::".join(components)
        return nodeid

    def process_docstring(self, docstring, node, nodeid=None, level=LEVELS.function):
        """Returns parsed Polarion docstring."""
        body = getattr(node, "body", None)
        body = body[0] if body else node

        # test doesn't have docstring, i.e. it's missing also the Polarion section
        if not docstring:
            return DocstringRecord(
                lineno=body.lineno - 1, column=node.col_offset, value={}, nodeid=nodeid, level=level
            )

        doc_list = docstring.split("\n")
        docstring_start = (
            body.lineno - 1 if hasattr(body, "end_lineno") else body.lineno - len(doc_list)
        )
        polarion_start, polarion_column = get_section_start(doc_list, self.SECTIONS.polarion)

        if not polarion_start:
            # docstring is missing the Polarion section
            return DocstringRecord(
                lineno=docstring_start + 1,
                column=node.col_offset + self.INDENT_STEP,
                value={},
                nodeid=nodeid,
                level=level,
            )
        # calculate polarion_column if the Polarion section is at the very begining of the docstring
        if polarion_start == 1 and polarion_column == node.col_offset:
            polarion_column += self.INDENT_STEP

        polarion_offset = docstring_start + polarion_start
        return DocstringRecord(
            lineno=polarion_offset,
            column=polarion_column,
            value=parse_docstring(docstring),
            nodeid=nodeid,
            level=level,
        )

    def process_node(self, node, class_name, level):
        """Returns parsed Polarion docstring present in the function."""
        docstring = self.get_docstring_from_node(node)
        nodeid = self._get_nodeid(getattr(node, "name", None), class_name)
        return self.process_docstring(docstring, node, nodeid, level)

    def process_ast_body(self, body, class_name=None, is_top=False):
        """Recursively iterates over specified part of ast tree to process functions."""
        docstrings = {}
        if is_top:
            docstring = self.process_node(body[0], None, LEVELS.file)
            if getattr(docstring, "value", None):
                docstrings[docstring.nodeid] = docstring
        for node in body:
            if isinstance(node, ast.ClassDef):
                if self.tests_only and not node.name.startswith("Test"):
                    continue
                docstring = self.process_node(node, None, LEVELS.klass)
                if getattr(docstring, "value", None):
                    docstrings[docstring.nodeid] = docstring
                docstrings.update(self.process_ast_body(node.body, node.name))
                continue

            if not isinstance(node, ast.FunctionDef):
                continue

            if self.tests_only and not node.name.startswith("test_"):
                continue

            docstring = self.process_node(node, class_name, LEVELS.function)
            docstrings[docstring.nodeid] = docstring

        return docstrings

    def get_docstrings(self):
        """Returns parsed Polarion docstrings present in the python source file."""
        return self.process_ast_body(self.tree.body, is_top=True)


# convenience functions


def parse_docstring(docstring):
    """Parses docstring to dictionary."""
    return DocstringParser(docstring).parse()


def strip_polarion_data(docstring):
    """Removes the Polarion section from the docstring."""
    return DocstringParser(docstring).strip_polarion_data()


def get_docstrings_in_file(filename, tree=None, tests_only=True):
    """Returns parsed Polarion docstrings present in the python source file."""
    return FileParser(filename, tree=tree, tests_only=tests_only).get_docstrings()


def _vrecords2dict(vrecord):
    vrecord = vrecord or {}
    converted = {}
    for key, value in vrecord.items():
        if key.startswith("_errors_"):
            continue
        converted[key] = [v.value for v in value] if isinstance(value, list) else value.value
    return converted


def _merge_components(docstrings_dict, nodeid):
    components = nodeid.split("::")

    merged_value = {}
    cur_components = []
    for component in components:
        cur_components.append(component)
        joined_components = "::".join(cur_components)
        merged_value.update(docstrings_dict.get(joined_components, {}))

    return merged_value


def merge_docstrings(docstrings):
    """Merges docstrings with their parents."""
    docstrings_dict = {}

    # create dict out of records
    for docstring in docstrings.values():
        if getattr(docstring, "value", False) is False:
            continue
        docstrings_dict[docstring.nodeid] = _vrecords2dict(docstring.value)

    # merge with parents
    for nodeid in docstrings_dict:
        # skip file-level docstrings and class docstrings
        if docstrings[nodeid].level != LEVELS.function:
            continue

        docstrings_dict[nodeid] = _merge_components(docstrings_dict, nodeid)

    return docstrings_dict


def merge_record(record, docstrings_dict, nodeid=None):
    """Merges single record with its parents based on nodeid."""
    merged_record = {}

    if isinstance(record, DocstringRecord):
        nodeid = nodeid or record.nodeid
        record = record.value

    if not nodeid:
        return merged_record

    merged_record = _merge_components(docstrings_dict, nodeid)
    merged_record.update(_vrecords2dict(record))
    return merged_record



"""
Checks Polarion docstrings.
"""

import os
import re
import sys
from collections import namedtuple

from polarion_tools_common import utils

from polarion_docstrings import parser

DocstringsError = namedtuple("DocstringsError", "lineno column message checker")
FieldRecord = namedtuple("FieldRecord", "lineno column field")
ValidatedDocstring = namedtuple("ValidatedDocstring", "unknown invalid missing markers ignored")

# regex for valid initialEstimate values
_NUM_INT = "[1-9][0-9]*"
_NUM_FRAC = "{nint}/{nint}".format(nint=_NUM_INT)
_NUM_INT_OR_FRAC = "{nint}(/{nint})?".format(nint=_NUM_INT)
INITIAL_ESTIMATE_RE = re.compile(
    "^(({num}d {num}h)|({nint} {frac}[dh])|{num}[dh])$".format(
        num=_NUM_INT_OR_FRAC, nint=_NUM_INT, frac=_NUM_FRAC
    )
)


# pylint: disable=too-many-instance-attributes
class DocstringsChecker:
    """Checker for Polarion docstrings."""

    PARSING_ERROR = "P663"
    IGNORED_FIELD = "P664"
    UNKNOWN_FIELD = "P666"
    INVALID_VALUE = "P667"
    MARKER_FIELD = "P668"
    MISSING_FIELD = "P669"

    def __init__(self, tree, filename, config, checker):
        self.tree = tree
        self.filename = filename
        self.config = self.get_valid_config(config)
        self.checker = checker or "DocstringsChecker"
        self.set_compiled_lists()

    def set_compiled_lists(self):
        """Sets compiled regular expressions for whitelist and blacklist.

        Uses values set in config if available so the compilation doesn't need to
        be repeated with every instantiation.
        """
        if "_compiled_blacklist" in self.config:
            self._compiled_whitelist = self.config.get("_compiled_whitelist")
            self._compiled_blacklist = self.config.get("_compiled_blacklist")
        else:
            self._compiled_whitelist, self._compiled_blacklist = self.get_compiled_lists(
                self.config
            )

    @staticmethod
    def get_valid_config(config):
        """Returns valid configuration if available."""
        cfg_valid = config.get("docstrings") or {}
        cfg_valid = cfg_valid.get("valid_values")
        if not (cfg_valid and config.get("default_fields")):
            return {}
        return config

    @staticmethod
    def get_compiled_lists(config):
        """Returns compiled regular expressions for whitelist and blacklist."""
        compiled_whitelist, compiled_blacklist = None, None
        if config.get("whitelisted_tests"):
            compiled_whitelist = re.compile("(" + ")|(".join(config.get("whitelisted_tests")) + ")")
        if config.get("blacklisted_tests"):
            compiled_blacklist = re.compile("(" + ")|(".join(config.get("blacklisted_tests")) + ")")
        return compiled_whitelist, compiled_blacklist

    @staticmethod
    def validate_value(docstring_dict, key, valid_values):
        record = docstring_dict.get(key)
        if record is not None:
            return record.value in valid_values[key]
        return True

    @staticmethod
    def get_unknown_fields(docstring_dict, known_fields):
        unknown = [
            FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
            for key in docstring_dict
            if key not in known_fields and not key.startswith("_errors_")
        ]
        return unknown

    @classmethod
    def get_invalid_fields(cls, docstring_dict, valid_values):
        results = {
            key: cls.validate_value(docstring_dict, key, valid_values) for key in valid_values
        }

        for section in (parser.DOCSTRING_SECTIONS.steps, parser.DOCSTRING_SECTIONS.results):
            # if "value" is present, the section wasn't parsed correctly into a list
            if hasattr(docstring_dict.get(section), "value"):
                results[section] = False

        invalid = [
            FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
            for key, result in results.items()
            if not result
        ]
        return invalid

    @staticmethod
    def get_missing_fields(docstring_record, required_keys, all_docstrings):
        """Look for required keys, and validate their value is not None"""
        missing = []

        if not required_keys:
            return missing

        # skip file-level docstrings and class docstrings
        if all_docstrings[docstring_record.nodeid].level != parser.LEVELS.function:
            return missing

        components = docstring_record.nodeid.split("::")
        all_components = [components.pop(0)]
        for component in components:
            joined_component = "{}::{}".format(all_components[-1], component)
            all_components.append(joined_component)
        all_components.reverse()

        for key in required_keys:
            value_record = None
            for nodeid in all_components:
                docstring_dict = all_docstrings.get(nodeid) or None
                docstring_dict = getattr(docstring_dict, "value", {})
                value_record = docstring_dict.get(key, None)
                if value_record is not None:
                    break
            if value_record is None or value_record.value is None:
                missing.append(key)

        return missing

    @staticmethod
    def get_markers_fields(docstring_dict, marker_fields):
        if not marker_fields:
            return []
        markers = [
            FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
            for key in marker_fields
            if key in docstring_dict
        ]
        return markers

    @staticmethod
    def get_ignored_fields(docstring_dict, ignored_fields):
        if not ignored_fields:
            return []
        ignored = [
            FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
            for key in ignored_fields
            if key in docstring_dict
        ]
        return ignored

    def validate_docstring(self, docstring_record, all_docstrings):
        """Returns tuple with lists of problematic fields."""
        docstring_dict = docstring_record.value or {}
        cfg_docstrings = self.config["docstrings"]
        unknown = self.get_unknown_fields(docstring_dict, self.config.get("default_fields"))
        invalid = self.get_invalid_fields(docstring_dict, cfg_docstrings.get("valid_values"))
        missing = self.get_missing_fields(
            docstring_record, cfg_docstrings.get("required_fields"), all_docstrings
        )
        markers = self.get_markers_fields(docstring_dict, cfg_docstrings.get("marker_fields"))
        ignored = self.get_ignored_fields(docstring_dict, cfg_docstrings.get("ignored_fields"))
        return ValidatedDocstring(unknown, invalid, missing, markers, ignored)

    def get_parsing_errors(self, docstring_dict, lineno=0):
        """Produces parsing errors for the flake8 checker."""
        docstring_dict = docstring_dict or {}
        if "_errors_parsing" not in docstring_dict:
            return []
        parsing_errors = [
            DocstringsError(
                lineno + err.lineno,
                err.column,
                "{} {}".format(self.PARSING_ERROR, err.message),
                self.checker,
            )
            for err in docstring_dict["_errors_parsing"]
        ]
        return parsing_errors

    # pylint:disable=too-many-locals
    def get_fields_errors(self, validated_docstring, docstring_dict, lineno=0, column=0):
        """Produces fields errors for the flake8 checker."""
        errors = []
        cfg_docstrings = self.config["docstrings"]
        marker_fields = cfg_docstrings.get("marker_fields") or {}
        ignored_fields = cfg_docstrings.get("ignored_fields") or {}

        for num, col, field in validated_docstring.unknown:
            errors.append(
                DocstringsError(
                    lineno + num,
                    col,
                    '{} Unknown field "{}"'.format(self.UNKNOWN_FIELD, field),
                    self.checker,
                )
            )
        for num, col, field in validated_docstring.invalid:
            errors.append(
                DocstringsError(
                    lineno + num,
                    col,
                    '{} Invalid value "{}" of the "{}" field'.format(
                        self.INVALID_VALUE, docstring_dict[field].value, field
                    ),
                    self.checker,
                )
            )
        for num, col, field in validated_docstring.markers:
            errors.append(
                DocstringsError(
                    lineno + num,
                    col,
                    '{} Field "{}" should be handled by the "{}" marker'.format(
                        self.MARKER_FIELD, field, marker_fields.get(field)
                    ),
                    self.checker,
                )
            )
        for num, col, field in validated_docstring.ignored:
            errors.append(
                DocstringsError(
                    lineno + num,
                    col,
                    '{} Ignoring field "{}": {}'.format(
                        self.IGNORED_FIELD, field, ignored_fields.get(field)
                    ),
                    self.checker,
                )
            )
        for field in validated_docstring.missing:
            errors.append(
                DocstringsError(
                    lineno,
                    column,
                    '{} Missing required field "{}"'.format(self.MISSING_FIELD, field),
                    self.checker,
                )
            )
        return errors

    def check_initial_estimate(self, docstring_dict, lineno=0):
        """Checks format of initialEstimate value."""
        initial_estimate = docstring_dict.get("initialEstimate")
        if not initial_estimate:
            return None

        if INITIAL_ESTIMATE_RE.sub("", initial_estimate.value):
            return DocstringsError(
                lineno + initial_estimate.lineno,
                initial_estimate.column,
                "{} Invalid estimate; "
                "Examples (d=days, h=hours): 1d, 3d, 1/2h, 2 1/2h, 3d 1h".format(
                    self.INVALID_VALUE
                ),
                self.checker,
            )
        return None

    @staticmethod
    def print_errors(errors):
        """Prints errors without using flake8."""
        for err in errors:
            print("line: {}:{}: {}".format(err.lineno, err.column, err.message), file=sys.stderr)

    def check_docstrings(self, docstrings_in_file):
        """Runs checks on each docstring."""
        errors = []
        for docstring in docstrings_in_file.values():
            if not self.is_nodeid_whitelisted(docstring.nodeid):
                continue
            valdoc = self.validate_docstring(docstring, docstrings_in_file)
            errors.extend(
                self.get_fields_errors(
                    valdoc, docstring.value, lineno=docstring.lineno, column=docstring.column
                )
            )
            errors.extend(self.get_parsing_errors(docstring.value, lineno=docstring.lineno))
            errors.append(self.check_initial_estimate(docstring.value, lineno=docstring.lineno))

        errors = [e for e in errors if e is not None]
        if errors:
            errors = sorted(errors, key=lambda tup: tup[0])
        return errors

    def run_checks(self):
        """Checks docstrings in python source file."""
        docstrings_in_file = parser.get_docstrings_in_file(self.filename, tree=self.tree)
        errors = self.check_docstrings(docstrings_in_file)
        return errors

    def is_nodeid_whitelisted(self, nodeid):
        """Checks if the nodeid is whitelisted."""
        if not nodeid:
            return True
        if self._compiled_whitelist and self._compiled_whitelist.search(nodeid):
            return True
        if self._compiled_blacklist and self._compiled_blacklist.search(nodeid):
            return False
        return True

    def is_file_for_check(self):
        """Decides if the file should be checked."""
        if not self.config:
            return False

        abs_filename = os.path.abspath(self.filename)
        head, tail = os.path.split(abs_filename)

        # check only test files under polarion tests path
        if not (tail.startswith("test_") and utils.find_tests_marker(head or ".")):
            return False

        return True

    def get_errors(self):
        """Get errors in docstrings in python source file."""
        if not self.is_file_for_check():
            return []

        return self.run_checks()



# pylint: disable=missing-docstring

from polarion_docstrings import checker, parser

EXPECTED = [
    (7, 4, 'P669 Missing required field "assignee"'),
    (7, 4, 'P669 Missing required field "initialEstimate"'),
    (11, 8, 'P669 Missing required field "assignee"'),
    (11, 8, 'P669 Missing required field "initialEstimate"'),
    (17, 8, 'P669 Missing required field "initialEstimate"'),
    (19, 12, 'P667 Invalid value "nonexistent" of the "casecomponent" field'),
    (40, 12, 'P667 Invalid value "level1" of the "caselevel" field'),
    (40, 12, 'P668 Field "caselevel" should be handled by the "@pytest.mark.tier" marker'),
    (41, 12, 'P668 Field "caseautomation" should be handled by the "@pytest.mark.manual" marker'),
    (
        42,
        12,
        'P668 Field "linkedWorkItems" should be handled by the "@pytest.mark.requirements" marker',
    ),
    (43, 12, 'P666 Unknown field "foo"'),
    (44, 12, 'P664 Ignoring field "description": use test docstring instead'),
    (51, 0, 'P669 Missing required field "assignee"'),
    (51, 0, 'P669 Missing required field "initialEstimate"'),
    (55, 0, 'P669 Missing required field "assignee"'),
    (55, 0, 'P669 Missing required field "initialEstimate"'),
    (61, 4, 'P669 Missing required field "assignee"'),
    (61, 4, 'P669 Missing required field "initialEstimate"'),
    (72, 8, 'P667 Invalid value "wrong" of the "testSteps" field'),
    (75, 8, 'P667 Invalid value "" of the "expectedResults" field'),
    (90, 4, 'P669 Missing required field "assignee"'),
    (91, 8, "P667 Invalid estimate; Examples (d=days, h=hours): 1d, 3d, 1/2h, 2 1/2h, 3d 1h"),
    (92, 12, "P663 Wrong indentation, line ignored"),
    (97, 4, 'P669 Missing required field "assignee"'),
    (98, 8, "P667 Invalid estimate; Examples (d=days, h=hours): 1d, 3d, 1/2h, 2 1/2h, 3d 1h"),
    (102, 15, "P663 Wrong indentation, line ignored"),
]


def _strip_func(errors):
    return [(lineno, col, msg) for lineno, col, msg, __ in errors]


def test_checker(source_file, config):
    errors = checker.DocstringsChecker(None, source_file, config, "TestChecker").run_checks()
    errors = _strip_func(errors)
    assert errors == EXPECTED


def test_initial_estimate(config):
    doc_checker = checker.DocstringsChecker(None, None, config, "TestChecker")

    valid = ("1d", "3d", "1/2h", "2 1/2h", "3d 1h", "3 2/5h")
    for value in valid:
        err = doc_checker.check_initial_estimate(
            {"initialEstimate": parser.ValueRecord(0, 0, value)}
        )
        assert err is None, "value {}".format(value)

    invalid = ("3h 1/60h", "3m", "1/60", "3d 01/60h", "3d 4d", "2 3h", "1/5 1/2h")
    for value in invalid:
        err = doc_checker.check_initial_estimate(
            {"initialEstimate": parser.ValueRecord(0, 0, value)}
        )
        assert "Invalid estimate" in err.message, "value {}".format(value)



"""
Checks Polarion docstrings using flake8.
"""

from pkg_resources import get_distribution
from polarion_tools_common import configuration

from polarion_docstrings import checker


def get_version():
    """Returns plugin version."""
    try:
        return get_distribution(__package__).version
    # pylint: disable=broad-except
    except Exception:
        # package is not installed
        return "0.0"


def set_compiled_lists(config):
    """Saves compiled regular expressions for whitelist and blacklist into config."""
    if not config:
        return
    config["_compiled_whitelist"], config[
        "_compiled_blacklist"
    ] = checker.DocstringsChecker.get_compiled_lists(config)


class PolarionDocstringsPlugin:
    """The flake8 entry point."""

    name = "polarion_checks"
    version = get_version()
    config = None

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = filename
        self.set_config(filename)

    @classmethod
    def set_config(cls, filename):
        """Sets plugin configuration.

        We set it once as a class attribute so the setup doesn't need to be done
        repeatedly for each checked file.
        """
        if cls.config is None:
            cls.config = configuration.get_config(project_path=filename) or {}
            set_compiled_lists(cls.config)

    def run(self):
        """Runs checks."""
        return checker.DocstringsChecker(
            self.tree, self.filename, self.config, PolarionDocstringsPlugin
        ).get_errors()



# pylint: disable=missing-docstring

from polarion_docstrings.parser import (
    DocstringRecord,
    ErrorRecord,
    ValueRecord,
    get_docstrings_in_file,
)

EXPECTED = {
    "tests/data/docstrings_func.py::TestClassFoo::test_in_class_no_docstring": DocstringRecord(
        lineno=7,
        column=4,
        value={},
        nodeid="tests/data/docstrings_func.py::TestClassFoo::test_in_class_no_docstring",
        level="function",
    ),
    "tests/data/docstrings_func.py::TestClassFoo::test_in_class_no_polarion": DocstringRecord(
        lineno=11,
        column=8,
        value={},
        nodeid="tests/data/docstrings_func.py::TestClassFoo::test_in_class_no_polarion",
        level="function",
    ),
    "tests/data/docstrings_func.py::TestClassFoo::test_in_class_polarion": DocstringRecord(
        lineno=17,
        column=8,
        value={
            "assignee": ValueRecord(lineno=1, column=12, value="mkourim"),
            "casecomponent": ValueRecord(lineno=2, column=12, value="nonexistent"),
            "caseimportance": ValueRecord(lineno=3, column=12, value="low"),
            "testSteps": [
                ValueRecord(
                    lineno=6,
                    column=16,
                    value="1. Step with really long description that doesn't fit into one line",
                ),
                ValueRecord(lineno=8, column=16, value="2. Do that"),
            ],
            "expectedResults": [
                ValueRecord(
                    lineno=11,
                    column=16,
                    value="1. Success outcome with really long description that doesn't "
                    "fit into one line",
                ),
                ValueRecord(lineno=13, column=16, value="2. second"),
            ],
            "title": ValueRecord(
                lineno=15,
                column=12,
                value="Some test with really long description that doesn't fit into one line",
            ),
            "setup": ValueRecord(
                lineno=18, column=12, value="Do this:\n- first thing\n- second thing"
            ),
            "teardown": ValueRecord(lineno=22, column=12, value="Tear it down"),
            "caselevel": ValueRecord(lineno=23, column=12, value="level1"),
            "caseautomation": ValueRecord(lineno=24, column=12, value="automated"),
            "linkedWorkItems": ValueRecord(lineno=25, column=12, value="FOO, BAR"),
            "foo": ValueRecord(lineno=26, column=12, value="this is an unknown field"),
            "description": ValueRecord(lineno=27, column=12, value="ignored"),
        },
        nodeid="tests/data/docstrings_func.py::TestClassFoo::test_in_class_polarion",
        level="function",
    ),
    "tests/data/docstrings_func.py::test_annotated_no_docstring": DocstringRecord(
        lineno=51,
        column=0,
        value={},
        nodeid="tests/data/docstrings_func.py::test_annotated_no_docstring",
        level="function",
    ),
    "tests/data/docstrings_func.py::test_standalone_no_docstring": DocstringRecord(
        lineno=55,
        column=0,
        value={},
        nodeid="tests/data/docstrings_func.py::test_standalone_no_docstring",
        level="function",
    ),
    "tests/data/docstrings_func.py::test_annotated_no_polarion": DocstringRecord(
        lineno=61,
        column=4,
        value={},
        nodeid="tests/data/docstrings_func.py::test_annotated_no_polarion",
        level="function",
    ),
    "tests/data/docstrings_func.py::test_annotated_polarion": DocstringRecord(
        lineno=69,
        column=4,
        value={
            "assignee": ValueRecord(lineno=1, column=8, value="mkourim"),
            "initialEstimate": ValueRecord(lineno=2, column=8, value="1/4h"),
            "testSteps": ValueRecord(lineno=3, column=8, value="wrong"),
            "expectedResults": ValueRecord(lineno=6, column=8, value=""),
        },
        nodeid="tests/data/docstrings_func.py::test_annotated_polarion",
        level="function",
    ),
    "tests/data/docstrings_func.py::test_blacklisted": DocstringRecord(
        lineno=82,
        column=4,
        value={"initialEstimate": ValueRecord(lineno=1, column=8, value="1/4h")},
        nodeid="tests/data/docstrings_func.py::test_blacklisted",
        level="function",
    ),
    "tests/data/docstrings_func.py::test_blacklisted_and_whitelisted": DocstringRecord(
        lineno=90,
        column=4,
        value={
            "initialEstimate": ValueRecord(lineno=1, column=8, value="1/4"),
            "_errors_parsing": [
                ErrorRecord(
                    lineno=2, column=12, type="indent", message="Wrong indentation, line ignored"
                )
            ],
        },
        nodeid="tests/data/docstrings_func.py::test_blacklisted_and_whitelisted",
        level="function",
    ),
    "tests/data/docstrings_func.py::test_wrong_steps_indent": DocstringRecord(
        lineno=97,
        column=4,
        value={
            "initialEstimate": ValueRecord(lineno=1, column=8, value="1/4m"),
            "testSteps": [
                ValueRecord(
                    lineno=3,
                    column=12,
                    value="1. Step with really long description that doesn't fit into one line",
                )
            ],
            "_errors_parsing": [
                ErrorRecord(
                    lineno=5, column=15, type="indent", message="Wrong indentation, line ignored"
                )
            ],
        },
        nodeid="tests/data/docstrings_func.py::test_wrong_steps_indent",
        level="function",
    ),
    "tests/data/docstrings_func.py::test_multiline": DocstringRecord(
        lineno=107,
        column=4,
        value={
            "assignee": ValueRecord(lineno=1, column=8, value="mkourim"),
            "initialEstimate": ValueRecord(lineno=2, column=8, value="1/4h"),
            "title": ValueRecord(
                lineno=3, column=8, value="Test multiple tenant quotas simultaneously"
            ),
            "testSteps": [
                ValueRecord(lineno=7, column=12, value="1."),
                ValueRecord(lineno=8, column=12, value="2. Log something"),
            ],
            "expectedResults": [
                ValueRecord(lineno=11, column=12, value="1."),
                ValueRecord(
                    lineno=12,
                    column=12,
                    value='2. <NoMethodError>: <undefined method `service_resources" for '
                    "nil:NilClass> [----] E, [2018-01-06T11:11:20.074019 #13027:e0ffc4] "
                    'ERROR -- :"',
                ),
            ],
        },
        nodeid="tests/data/docstrings_func.py::test_multiline",
        level="function",
    ),
}


def test_parser(source_file):
    docstrings = get_docstrings_in_file(source_file)
    assert docstrings == EXPECTED



# pylint: disable=missing-docstring,no-self-use
"""Test file

Polarion:
    assignee: mkourim
    foo: this is an unknown field
"""

import pytest


class TestClassFoo:
    """Test class

    Polarion:
        assignee: psegedy
        caseimportance: huge
        bar: this is an unknown field
        linkedWorkItems: FOO, BAR
            initialEstimate: 1/4h

        testSteps:
            1. Step with really long description
               that doesn't fit into one line
            2. Do that
    """

    def test_in_class_no_docstring(self):
        pass

    def test_in_class_no_polarion(self):
        """FOO"""

    @pytest.mark.skip
    def test_in_class_polarion(self):
        """FOO

        Polarion:
            assignee: mkourim
            caseimportance: low
            baz: this is an unknown field

            testSteps:
                1. Do that
        """


def test_standalone_no_docstring():
    pass



# pylint: disable=missing-docstring

from setuptools import find_packages, setup

with open("README.rst", "rb") as fp:
    LONG_DESCRIPTION = fp.read().decode("utf-8").strip()

setup(
    name="polarion-docstrings",
    use_scm_version=True,
    url="https://gitlab.com/mkourim/polarion_docstrings",
    description="Reads Polarion docstrings and validates them using flake8",
    long_description=LONG_DESCRIPTION,
    author="Martin Kourim",
    author_email="mkourim@redhat.com",
    license="MIT",
    packages=find_packages(exclude=("tests",)),
    setup_requires=["setuptools_scm"],
    install_requires=["pyyaml", "polarion-tools-common"],
    entry_points={
        "flake8.extension": ["P66 = polarion_docstrings.flake8_plugin:PolarionDocstringsPlugin"]
    },
    keywords=["polarion", "testing"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Testing",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Developers",
    ],
    include_package_data=True,
)
